#include "StdAfx.h"
#include "JSPluginManager.h"
#include "../js/JSValueWrapper.h"
#include "../window/IWindow.h"
#include "../js/JSDispatch.h"
#include "../js/IJSContext.h"
#include "../ar/ARWindow.h"
#include "../common/URLPath.h"

#define FUNC_LOAD "load"
#define PROP_PLUGIN_COUNT	"pluginCount"

#define DLL_FUNCTION_CREATE_PLUGIN "_CreatePlugin@8"

// \z܂B
JSPluginManager::JSPluginManager(const char* basePath, IWindow* pWindow, ARWindow* pARWindow)
: JSMemberCollection("JSPluginManager")
{
	m_basePath = basePath;
	m_pWindow = pWindow;
	m_pARWindow = pARWindow;
	m_pJSEngineContext = NULL;
	m_pEnv = new ARGPluginEnvironmentImpl(this);
}

// j܂B
JSPluginManager::~JSPluginManager()
{
	for(vector<IARGPlugin*>::size_type i = 0; i < m_plugins.size(); i ++) {
		IARGPlugin* pPlugin = m_plugins[i].second;
		if(pPlugin != NULL) {
			pPlugin->Detach();
			pPlugin->Release();
			pPlugin = NULL;
		}

		HMODULE hModule = m_plugins[i].first;
		if(hModule != NULL) {
			FreeLibrary(hModule);
			hModule = NULL;
		}

		m_plugins[i].first = NULL;
		m_plugins[i].second = NULL;
	}
	m_plugins.clear();
	if(m_pEnv != NULL) {
		m_pEnv->Release();
		m_pEnv = NULL;
	}
}

// j܂B
void JSPluginManager::Destroy()
{
	delete this;
}

// JavaScriptGWݒ肵܂B
HRESULT JSPluginManager::SetJSEngineContext(JSEngineContext* pContext)
{
	if(pContext == NULL) {
		return E_POINTER;
	}
	m_pJSEngineContext = pContext;
	return S_OK;
}

// o܂B
HRESULT JSPluginManager::InitMembers()
{
	HRESULT hr = S_OK;

	hr = DefineFunction(FUNC_LOAD, 2);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty(PROP_PLUGIN_COUNT, true);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

typedef HRESULT (__stdcall *CreatePluginFunc) (const char*, IUnknown** ppUnk);

// ֐s܂B
HRESULT JSPluginManager::InvokeFunction(JSContext *cx, JSObject *obj, const char* name, uintN argc, jsval *argv, jsval *rval)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == FUNC_LOAD) {
		if(argc < 2) {
			return E_INVALIDARG;
		}
		
		string path;
		JSValueWrapper value1(cx, argv[0]);
		HRESULT hr = value1.GetString(path);
		if(FAILED(hr)) {
			return hr;
		}

		string name;
		JSValueWrapper value2(cx, argv[1]);
		hr = value2.GetString(name);
		if(FAILED(hr)) {
			return hr;
		}

		string fullpath = m_basePath + path;
		HMODULE hModule = LoadLibraryA(fullpath.c_str());
		if(hModule == NULL) {
			return E_UNEXPECTED;
		}

		CreatePluginFunc pFunc = NULL;
		pFunc = (CreatePluginFunc) GetProcAddress(hModule, DLL_FUNCTION_CREATE_PLUGIN);
		if(pFunc == NULL) {
			return E_UNEXPECTED;
		}

		IUnknown* pUnk = NULL;
		hr = pFunc(name.c_str(), & pUnk);
		if(FAILED(hr)) {
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}
		if(pUnk == NULL) {
			FreeLibrary(hModule);
			hModule = NULL;
			return E_UNEXPECTED;
		}

		IARGPlugin* pPlugin = NULL;
		hr = pUnk->QueryInterface(IID_IARGPlugin, (void**) & pPlugin);
		if(FAILED(hr)) {
			pUnk->Release();
			pUnk = NULL;
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}
		pUnk->Release();
		pUnk = NULL;

		hr = pPlugin->Attach(m_pEnv);
		if(FAILED(hr)) {
			pPlugin->Release();
			pPlugin = NULL;
			FreeLibrary(hModule);
			hModule = NULL;
			return hr;
		}

		pair<HMODULE,IARGPlugin*> loaded(hModule, pPlugin);
		m_plugins.push_back(loaded);

		return S_OK;
	}
	return JSMemberCollection::InvokeFunction(cx, obj, name, argc, argv, rval);
}

// vpeB擾܂B
HRESULT JSPluginManager::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(cx == NULL || obj == NULL || name == NULL) {
		return E_POINTER;
	}
	string sname = name;
	if(sname == PROP_PLUGIN_COUNT) {
		JSValueWrapper value(cx, JSVAL_NULL);
		HRESULT hr = value.SetInt(m_plugins.size());
		if(FAILED(hr)) {
			return hr;
		}

		return value.Get(vp);
	}
	return JSMemberCollection::GetProperty(cx, obj, name, vp);
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::ARGPluginEnvironmentImpl(JSPluginManager* pOwner)
{
	m_pOwner = pOwner;
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::~ARGPluginEnvironmentImpl()
{
	m_pOwner = NULL;
}

// pX擾܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::GetRealPath(const char* relativePath, char* absolutePath, int maxLen, int* pLen)
{
	if(relativePath == NULL || pLen == NULL) {
		return E_POINTER;
	}

	string url;
	HRESULT hr = m_pOwner->m_pContext->GetAbsoluteURL(relativePath, url);
	if(FAILED(hr)) {
		return hr;
	}

	URLPath urlPath(url.c_str(), & hr);
	if(FAILED(hr)) {
		return hr;
	}
	string realFile;
	hr = urlPath.GetRealPath(realFile);
	if(FAILED(hr)) {
		return hr;
	}

	if(absolutePath != NULL) {
		strcpy_s(absolutePath, maxLen, realFile.c_str());
		if(realFile.length() < maxLen) {
			absolutePath[realFile.length()] = '\0';
		}
	}

	*pLen = realFile.length();

	return S_OK;
}

// vpeBǉ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::AddProperty(const char* name, IDispatch* pDisp)
{
	if(name == NULL || pDisp == NULL) {
		return E_POINTER;
	}
	if(m_pOwner->m_pJSEngineContext == NULL) {
		return E_UNEXPECTED;
	}

	HRESULT hr = S_OK;
	JSDispatch* pJSDisp = new JSDispatch(pDisp, & hr);
	if(FAILED(hr)) {
		delete pJSDisp;
		return hr;
	}

	hr = pJSDisp->Attach(m_pOwner->m_pJSEngineContext, NULL);
	if(FAILED(hr)) {
		return hr;
	}
	return m_pOwner->m_pJSEngineContext->DefineGlobalProperty(name, pJSDisp);
}

// IuWFNgt@Ngǉ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::AddObjectFactory(const char* name, IARGObjectFactory* pObjectFactory)
{
	if(name == NULL || pObjectFactory == NULL) {
		return E_POINTER;
	}
	return m_pOwner->m_pARWindow->AddObjectFactory(name, pObjectFactory);
}

// p^[oǉ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::AddPatternDetector(const char* name, IARGPatternDetector* pPatternDetector)
{
	if(name == NULL || pPatternDetector == NULL) {
		return E_POINTER;
	}
	return m_pOwner->m_pARWindow->AddPatternDetector(name, new PatternDetectorWrapper(m_pOwner->m_pJSEngineContext, pPatternDetector));
}

// R\[ɕo͂܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::PrintConsole(const char* text)
{
	if(text == NULL) {
		return E_POINTER;
	}
	return m_pOwner->m_pWindow->PrintLine(text);
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::PatternDetectorWrapper(JSEngineContext* pContext, IARGPatternDetector* pDetector)
{
	m_pContext = pContext;
	m_pDetector = pDetector;
	pDetector->AddRef();
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::~PatternDetectorWrapper()
{
	if(m_pDetector != NULL) {
		m_pDetector->Release();
		m_pDetector = NULL;
	}
	for(vector<JSPattern*>::size_type i = 0; i < m_patterns.size(); i ++) {
		JSPattern* pDisp = m_patterns[i];
		pDisp->Destroy();
		pDisp = NULL;
	}
	m_patterns.clear();
}

// j܂B
void JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::Destroy()
{
	delete this;
}

// ̃p[^擾܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::GetCreateParamCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	int candCount = 0;
	HRESULT hr = m_pDetector->GetParameterTypeCandidateCount(& candCount);
	if(FAILED(hr)) {
		return hr;
	}

	int minCount = -1;
	for(int i = 0; i < candCount; i ++) {
		int paramCount = 0;
		hr = m_pDetector->GetParameterTypeCount(i, & paramCount);
		if(FAILED(hr)) {
			return hr;
		}
		if(minCount < 0) {
			minCount = paramCount;
		}else if(paramCount < minCount) {
			minCount = paramCount;
		}
	}
	if(minCount < 0) {
		return E_UNEXPECTED;
	}
	*pCount = minCount;
	return S_OK;
}

// p^[ǉ܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::Create(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	int candCount = 0;
	HRESULT hr = m_pDetector->GetParameterTypeCandidateCount(& candCount);
	if(FAILED(hr)) {
		return hr;
	}

	for(int i = 0; i < candCount; i ++) {
		int paramCount = 0;
		hr = m_pDetector->GetParameterTypeCount(i, & paramCount);
		if(FAILED(hr)) {
			return hr;
		}
		if(paramCount == argc) {
			VARIANT* pVars = NULL;
			if(argc == 0) {
				pVars = new VARIANT[1];
			}else{
				pVars = new VARIANT[argc];
			}
			for(int j = 0; j < paramCount; j ++) {
				VariantInit(pVars + j);
			}

			for(int j = 0; j < paramCount; j ++) {
				VARTYPE vt = VT_NULL;
				hr = m_pDetector->GetParameterType(i, j, & vt);
				if(FAILED(hr)) {
					return hr;
				}

				hr = JSDispatch::ConvertJSVALtoVARIANT(m_pContext, argv[j], pVars + j, vt);
				if(FAILED(hr)) {
					return hr;
				}
			}

			IARGPattern* pObj = NULL;
			hr = m_pDetector->Create(pVars, argc, & pObj);
			if(FAILED(hr)) {
				return hr;
			}

			hr = S_OK;
			JSPattern* pJSDisp = new JSPattern(pObj, & hr);
			if(FAILED(hr)) {
				return hr;
			}

			m_patterns.push_back(pJSDisp);

			pObj->Release();
			pObj = NULL;

			for(int j = 0; j < paramCount; j ++) {
				VariantClear(pVars + j);
			}
			delete[] pVars;
			pVars = NULL;

			hr = pJSDisp->Attach(m_pContext, NULL);
			if(FAILED(hr)) {
				return hr;
			}

			if(rval == NULL) {
				return S_OK;
			}
			*rval = OBJECT_TO_JSVAL(pJSDisp->GetCore());
			return S_OK;
		}
	}
	return E_UNEXPECTED;
}

// osȂ܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::Process(const CAPTURE_IMAGE* pImage)
{
	return m_pDetector->Process(pImage->width, pImage->height, pImage->pBuffer);
}

// p^[擾܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::GetCount(int* pCount)
{
	if(pCount == NULL) {
		return E_POINTER;
	}
	*pCount = (int) m_patterns.size();
	return S_OK;
}

// p^[擾܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::Get(int index, IJSObject** ppObject)
{
	if(ppObject == NULL) {
		return E_POINTER;
	}
	JSPattern* pPat = m_patterns[index];
	*ppObject = pPat;
	return S_OK;
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSPattern(IARGPattern* pPattern, HRESULT* phr)
: JSMemberCollection("JSPattern")
{
	m_pPattern = pPattern;
	m_pParam = NULL;
	m_pJSEvent = NULL;

	if(pPattern == NULL) {
		*phr = E_POINTER;
		return;
	}
	pPattern->AddRef();
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::~JSPattern()
{
	if(m_pPattern != NULL) {
		m_pPattern->Release();
		m_pPattern = NULL;
	}
	if(m_pParam != NULL) {
		m_pParam->Destroy();
		m_pParam = NULL;
	}
	if(m_pJSEvent != NULL){
		m_pJSEvent->Destroy();
		m_pJSEvent = NULL;
	}
}

// j܂B
void JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::Destroy()
{
	delete this;
}

// ܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::InitMembers()
{
	ObserverImpl* pObserver = new ObserverImpl(this);
	HRESULT hr = m_pPattern->AddPatternObserver(pObserver);
	if(FAILED(hr)) {
		return hr;
	}
	pObserver->Release();
	pObserver = NULL;

	hr = DefineProperty("translate", true);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("params", true);
	if(FAILED(hr)) {
		return hr;
	}

	hr = DefineProperty("onfound", false);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("onlost", false);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("onchanged", false);
	if(FAILED(hr)) {
		return hr;
	}

	hr = m_onFound.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}
	hr = m_onLost.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}
	hr = m_onChanged.Init(m_pContext, this);
	if(FAILED(hr)) {
		return hr;
	}

	return S_OK;
}

// vpeBݒ肵܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::SetProperty(JSContext *cx, JSObject *obj, const char* name, jsval vp)
{
	string sname = name;
	if(sname == "onfound") {
		return m_onFound.Set(vp);
	}else if(sname == "onlost") {
		return m_onLost.Set(vp);
	}else if(sname == "onchanged") {
		return m_onChanged.Set(vp);
	}
	return JSMemberCollection::SetProperty(cx, obj, name, vp);
}

// vpeB擾܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	string sname = name;
	if(sname == "translate") {
		PATTERN_TRANSLATE trans;
		ZeroMemory(& trans, sizeof(trans));
		HRESULT hr = m_pPattern->GetTranslate(& trans);
		if(FAILED(hr)) {
			return hr;
		}
		if(hr == S_FALSE) {
			*vp = JSVAL_NULL;
			return S_OK;
		}
		

		jsval transValues[3];
		for(int i = 0; i < 3; i ++) {
			JSObject* pArray = NULL;
			HRESULT hr = CreateDoubleArray(trans.matrix[i], 4, & pArray);
			if(FAILED(hr)) {
				return hr;
			}
			transValues[i] = OBJECT_TO_JSVAL(pArray);
		}
		JSObject* pTransObj = JS_NewArrayObject(cx, 3, transValues);
		if(pTransObj == NULL) {
			return E_FAIL;
		}
		*vp = OBJECT_TO_JSVAL(pTransObj);
		return S_OK;
	}else if(sname == "params") {
		if(m_pParam == NULL) {
			IDispatch* pDisp = NULL;
			HRESULT hr = m_pPattern->QueryInterface(& pDisp);
			if(hr == E_NOINTERFACE) {
				*vp = JSVAL_NULL;
				return S_OK;
			}
			if(FAILED(hr)) {
				return hr;
			}

			hr = S_OK;
			m_pParam = new JSDispatch(pDisp, & hr);
			if(FAILED(hr)) {
				delete m_pParam;
				m_pParam = NULL;
				return hr;
			}

			hr = m_pParam->Attach(m_pContext, NULL);
			if(FAILED(hr)) {
				delete m_pParam;
				m_pParam = NULL;
				return hr;
			}
			if(pDisp != NULL) {
				pDisp->Release();
				pDisp = NULL;
			}
		}

		*vp = OBJECT_TO_JSVAL(m_pParam->GetCore());
		return S_OK;
	}else if(sname == "onfound") {
		return m_onFound.Get(vp);
	}else if(sname == "onlost") {
		return m_onLost.Get(vp);
	}else if(sname == "onchanged") {
		return m_onChanged.Get(vp);
	}
	return JSMemberCollection::GetProperty(cx, obj, name, vp);
}

// ύXs܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::DispatchChanged(PATTERN_STATE state)
{
	HRESULT hr = S_OK;
	JSObject* pEvent = NULL;
	hr = CreatePatternEvent(state, & pEvent);
	if(FAILED(hr)) {
		return hr;
	}

	FunctionRef* pFunc = NULL;
	if(state == PATTERN_FOUND) {
		pFunc = & m_onFound;
	}else if(state == PATTERN_LOST) {
		pFunc = & m_onLost;
	}else if(state == PATTERN_CHANGED) {
		pFunc = & m_onChanged;
	}
	if(pFunc == NULL) {
		return S_OK;
	}

	jsval eventVal = OBJECT_TO_JSVAL(pEvent);
	jsval rval = JSVAL_FALSE;
	hr = pFunc->InvokeSync(& eventVal, 1, & rval);
	if(FAILED(hr)) {
		return hr;
	}
	return S_OK;
}

// p^[Cxg𐶐܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::CreatePatternEvent(PATTERN_STATE state, JSObject** ppEvent)
{
	if(ppEvent == NULL) {
		return E_POINTER;
	}

	if(m_pJSEvent == NULL) {
		HRESULT hr = S_OK;
		m_pJSEvent = new JSEventImpl(this, state, & hr);
		if(m_pJSEvent == NULL) {
			return E_OUTOFMEMORY;
		}
		if(FAILED(hr)) {
			return hr;
		}
		hr = m_pJSEvent->Attach(m_pContext, NULL);
		if(FAILED(hr)) {
			return hr;
		}
	}else{
		HRESULT hr = m_pJSEvent->Set(state);
		if(FAILED(hr)) {
			return hr;
		}
	}

	*ppEvent = m_pJSEvent->GetCore();
	return S_OK;
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::ObserverImpl::ObserverImpl(JSPattern* pOwner)
{
	m_pOwner = pOwner;
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::ObserverImpl::~ObserverImpl()
{
	m_pOwner = NULL;
}

// C^tF[X擾܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::ObserverImpl::QueryInterface(const IID& riid, void** ppv)
{
	if(ppv == NULL) {
		return E_POINTER;
	}

	if(riid == IID_IARGPatternObserver) {
		*ppv = static_cast<IARGPatternObserver*>(this);
		AddRef();
		return S_OK;
	}

	return ARGBaseUnknown::QueryInterface(riid, ppv);
}

// Cxg̏sȂ܂B
STDMETHODIMP JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::ObserverImpl::DispatchChanged(PATTERN_STATE state)
{
	return m_pOwner->DispatchChanged(state);
}

// \z܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::JSEventImpl(JSPattern* pOwner, PATTERN_STATE state, HRESULT* phr)
: JSMemberCollection("JSPattern::JSEventImpl")
{
	m_pOwner = pOwner;
	m_state = PATTERN_LOST;
	*phr = S_OK;

	HRESULT hr = Set(state);
	if(FAILED(hr)) {
		*phr = hr;
		return;
	}
}

// j܂B
JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::~JSEventImpl()
{
	m_pOwner = NULL;
}

// j܂B
void JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::Destroy()
{
	delete this;
}

// Ԃݒ肵܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::Set(PATTERN_STATE state)
{
	m_state = state;
	return S_OK;
}

// ܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::InitMembers()
{
	HRESULT hr = S_OK;

	hr = DefineProperty("target", true);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("found", true);
	if(FAILED(hr)) {
		return hr;
	}
	hr = DefineProperty("translate", true);
	if(FAILED(hr)) {
		return hr;
	}
	return S_OK;
}

// vpeB擾܂B
HRESULT JSPluginManager::ARGPluginEnvironmentImpl::PatternDetectorWrapper::JSPattern::JSEventImpl::GetProperty(JSContext* cx, JSObject* obj, const char* name, jsval *vp)
{
	if(vp == NULL) {
		return S_OK;
	}
	string sname = name;
	if(sname == "target") {
		jsval targetValue = OBJECT_TO_JSVAL(m_pOwner->GetCore());
		*vp = targetValue;
		return S_OK;
	}else if(sname == "found") {
		JSValueWrapper valuew(cx, JSVAL_NULL);

		bool found = false;
		if(m_state != PATTERN_LOST) {
			found = true;
		}
		HRESULT hr = valuew.SetBool(found);
		if(FAILED(hr)) {
			return hr;
		}
		return valuew.Get(vp);
	}else if(sname == "translate") {
		if(m_state == PATTERN_LOST) {
			*vp = JSVAL_NULL;
			return S_OK;
		}

		PATTERN_TRANSLATE trans;
		ZeroMemory(& trans, sizeof(trans));
		HRESULT hr = m_pOwner->m_pPattern->GetTranslate(& trans);
		if(FAILED(hr)) {
			return hr;
		}

		jsval transValues[3];
		for(int i = 0; i < 3; i ++) {
			JSObject* pArray = NULL;
			hr = CreateDoubleArray(trans.matrix[i], 4, & pArray);
			if(FAILED(hr)) {
				return hr;
			}
			transValues[i] = OBJECT_TO_JSVAL(pArray);
		}
		JSObject* pTransObj = JS_NewArrayObject(cx, 3, transValues);
		if(pTransObj == NULL) {
			return E_FAIL;
		}
		*vp = OBJECT_TO_JSVAL(pTransObj);
		return S_OK;
	}else{
		return E_NOTIMPL;
	}
}
